#!/usr/bin/python3
"""
## © University of Southampton IT Innovation Centre, 2018
##
## Copyright in this software belongs to University of Southampton
## IT Innovation Centre of Gamma House, Enterprise Road,
## Chilworth Science Park, Southampton, SO16 7NS, UK.
##
## This software may not be used, sold, licensed, transferred, copied
## or reproduced in whole or in part in any manner or form or in or
## on any media by any person other than in accordance with the terms
## of the Licence Agreement supplied with the software, or otherwise
## without the prior written consent of the copyright owners.
##
## This software is distributed WITHOUT ANY WARRANTY, without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
## PURPOSE, except where stated in the Licence Agreement supplied with
## the software.
##
##      Created By :            Nikolay Stanchev
##      Created Date :          25-04-2018
##      Created for Project :   FLAME
"""

from threading import Thread, Event
from influxdb import InfluxDBClient
from time import time, sleep
from urllib.parse import urlparse
from clmcservice.utilities import generate_e2e_delay_report
import getopt
import logging


class Aggregator(object):
    """
    A class used to perform the aggregation feature of the CLMC - aggregating network and media service measurements. Implemented as a separate process.
    """

    REPORT_PERIOD = 5  # default report period is 5s, that is every 5 seconds the mean delay values for the last 5 seconds are aggregated
    DATABASE = 'CLMCMetrics'  # default database the aggregator uses
    DATABASE_URL = 'http://172.40.231.51:8086'  # default database URL the aggregator uses

    def __init__(self, database_name=DATABASE, database_url=DATABASE_URL, report_period=REPORT_PERIOD, logger=None):
        """
        Constructs an Aggregator instance.

        :param database_name: database name to use
        :param database_url: database url to use
        :param report_period: the report period in seconds
        """

        if logger is None:
            self.log = logging.getLogger(__name__)
        else:
            self.log = logger

        self.log.info("Connecting to Influx database {0} with URL {1}".format(database_name, database_url))
        # initialise a database client using the database url and the database name
        url_object = urlparse(database_url)
        self.db_client = InfluxDBClient(host=url_object.hostname, port=url_object.port, database=database_name, timeout=10)
        self.log.info("Successfully connected to Influx database {0} with URL {1}".format(database_name, database_url))

        self.db_url = database_url
        self.db_name = database_name
        self.report_period = report_period

        # a cache-like dictionaries to store the last reported values, which can be used to fill in missing values
        self.network_cache = {}
        self.service_cache = {}

        # a stop flag event object used to handle the stopping of the process
        self._stop_flag = Event()

    def stop(self):
        """
        Stop the aggregator from running.
        """

        self.log.info("Aggregator's stop flag has been set.")
        self._stop_flag.set()

    def run(self):
        """
        Performs the functionality of the aggregator - query data from both measurements merge that data and post it back in influx every 5 seconds.
        """

        self.log.info("Aggregator started running.")

        current_time = int(time())
        while not self._stop_flag.is_set():
            self.log.info("Trying to generate an E2E measurement.")

            boundary_time = current_time - self.report_period

            boundary_time_nano = boundary_time * 1000000000
            current_time_nano = current_time * 1000000000

            # query the network delays and group them by path ID
            network_delays = {}
            result = self.db_client.query(
                'SELECT mean(latency) as "net_latency", mean(bandwidth) as "net_bandwidth" FROM "{0}"."autogen"."network_delays" WHERE time >= {1} and time < {2} GROUP BY path, source, target'.format(
                    self.db_name, boundary_time_nano, current_time_nano))

            for item in result.items():
                metadata, result_points = item
                # measurement = metadata[0]
                tags = metadata[1]

                result_point = next(result_points)
                network_delays[(tags['path'], tags['source'], tags['target'])] = result_point['net_latency'], result_point['net_bandwidth']
                self.network_cache[(tags['path'], tags['source'], tags['target'])] = result_point['net_latency'], result_point['net_bandwidth']

            # query the service delays and group them by endpoint, service function instance and sfr
            service_delays = {}
            result = self.db_client.query(
                'SELECT mean(response_time) as "response_time", mean(request_size) as "request_size", mean(response_size) as "response_size" FROM "{0}"."autogen"."service_delays" WHERE time >= {1} and time < {2} GROUP BY endpoint, sf_instance, sfr'.format(
                    self.db_name, boundary_time_nano, current_time_nano))

            for item in result.items():
                metadata, result_points = item
                # measurement = metadata[0]
                tags = metadata[1]
                result_point = next(result_points)
                service_delays[tags['sfr']] = (result_point['response_time'], result_point['request_size'], result_point['response_size'], tags['endpoint'], tags['sf_instance'])
                self.service_cache[tags['sfr']] = (result_point['response_time'], result_point['request_size'], result_point['response_size'], tags['endpoint'], tags['sf_instance'])

            # for each network path check if there is a media service delay report for the target sfr - if so, generate an e2e_delay measurement
            for path in network_delays:
                path_id, source, target = path

                # check if we have a reverse path without a forward path for a potential aggregated row - e.g. SR3 to SR1 network row with service on SR3 and no row from SR1 to SR3
                if (source in service_delays or source in self.service_cache) and (path_id, target, source) not in network_delays and (path_id, target, source) in self.network_cache:
                    # hence search for the forward path in the cache
                    forward_path = self.network_cache.get((path_id, target, source))
                    reverse_path = network_delays.get((path_id, source, target))
                    forward_delay = forward_path[0]
                    avg_bandwidth = forward_path[1]
                    reverse_delay = reverse_path[0]
                    service_delay = service_delays.get(source, self.service_cache.get(source))
                    response_time, request_size, response_size, endpoint, sf_instance = service_delay
                    self.db_client.write_points(
                        generate_e2e_delay_report(path_id, target, source, endpoint, sf_instance, forward_delay, reverse_delay, response_time,
                                                  request_size, response_size, avg_bandwidth, boundary_time))

                # check if target sfr is reported in service delays, in other words - if there is a media service instance being connected to target sfr
                if target not in service_delays and target not in self.service_cache:
                    # if not continue with the other network path reports
                    continue

                e2e_arguments = {"path_ID": None, "source_SFR": None, "target_SFR": None, "endpoint": None, "sf_instance": None, "delay_forward": None, "delay_reverse": None,
                                 "delay_service": None, "avg_request_size": None, "avg_response_size": None, "avg_bandwidth": None, "time": boundary_time}

                e2e_arguments['path_ID'] = path_id
                e2e_arguments['source_SFR'] = source
                e2e_arguments['target_SFR'] = target
                e2e_arguments['delay_forward'] = network_delays[path][0]
                e2e_arguments['avg_bandwidth'] = network_delays[path][1]

                # reverse the path ID to get the network delay for the reversed path
                reversed_path = (path_id, target, source)
                if reversed_path in network_delays or reversed_path in self.network_cache:
                    # get the reverse delay, use the latest value if reported or the cache value
                    e2e_arguments['delay_reverse'] = network_delays.get(reversed_path, self.network_cache.get(reversed_path))[0]
                else:
                    e2e_arguments['delay_reverse'] = None

                # get the response time of the media component connected to the target SFR
                service_delay = service_delays.get(target, self.service_cache.get(target))
                response_time, request_size, response_size, endpoint, sf_instance = service_delay
                # put these points in the e2e arguments dictionary
                e2e_arguments['delay_service'] = response_time
                e2e_arguments['avg_request_size'] = request_size
                e2e_arguments['avg_response_size'] = response_size
                e2e_arguments['endpoint'] = endpoint
                e2e_arguments['sf_instance'] = sf_instance

                # if all the arguments of the e2e delay measurements were reported, then generate and post to Influx an E2E measurement row
                if None not in e2e_arguments.values():
                    self.db_client.write_points(
                        generate_e2e_delay_report(e2e_arguments['path_ID'], e2e_arguments['source_SFR'], e2e_arguments['target_SFR'], e2e_arguments['endpoint'],
                                                  e2e_arguments['sf_instance'], e2e_arguments['delay_forward'], e2e_arguments['delay_reverse'],
                                                  e2e_arguments['delay_service'],
                                                  e2e_arguments["avg_request_size"], e2e_arguments['avg_response_size'], e2e_arguments['avg_bandwidth'],
                                                  e2e_arguments['time']))
                    self.log.info("Successfully generated an E2E measurement and posted back to Influx.")
                else:
                    self.log.info("Couldn't generate an E2E measurement although some of the data could be fetched.")

            old_timestamp = current_time
            # wait until {report_period) seconds have passed
            while current_time < old_timestamp + self.report_period:
                sleep(1)
                current_time = int(time())

        self.log.info("Aggregator stopped running.")


class AggregatorThread(Thread):
    """
    A utility class used to wrap around the Aggregator class and return a Thread instance, which can then be used for testing (provides start and stop methods)
    """

    REPORT_PERIOD = 5  # currently, report period is 5s, that is every 5 seconds the mean delay values for the last 5 seconds are aggregated
    DATABASE = 'CLMCMetrics'  # default database the aggregator uses
    DATABASE_URL = 'http://172.40.231.51:8086'  # default database URL the aggregator uses

    def __init__(self, database=DATABASE, database_url=DATABASE_URL, report_period=REPORT_PERIOD):
        """
        Constructs an Aggregator instance.

        :param database: database name to use
        :param database_url: database url to use
        """

        super(AggregatorThread, self).__init__()  # call the constructor of the thread

        self.aggregator = Aggregator(database_name=database, database_url=database_url, report_period=report_period)

    def stop(self):
        """
        A method used to stop the thread.
        """

        self.aggregator.stop()

    def add_event_lock(self, event):
        """
         Auxiliary method to set a thread-safe event lock object to the aggregator (used for testing).

         :param event: the event lock object
         """

        setattr(self, 'event', event)

    def run(self):
        """
        The method to execute when the thread starts.
        """

        if hasattr(self, 'event'):
            self.event.set()

        self.aggregator.run()


if __name__ == '__main__':
    # initialise a file logger, only when module's main method is run (NOT when aggregator class is imported somewhere else)
    log = logging.getLogger('aggregator')
    hdlr = logging.FileHandler('/var/log/flame/clmc/aggregator.log', mode='a')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    log.addHandler(hdlr)
    log.setLevel(logging.DEBUG)

    # log all errors that are thrown in the execution of the aggregator with the logger object initialized above
    import sys
    import traceback

    def report_error(error_type, error_value, error_traceback, log_object=log):
        log_object.error("Uncaught error thrown!")
        log_object.error("Error type: {0}".format(error_type))
        log_object.error("Error value: {0}".format(error_value))
        log_object.debug("Error traceback:")
        for trace in traceback.format_tb(error_traceback):
            log_object.debug(trace)

    sys.excepthook = report_error

    # Parse command line options
    try:
        opts, args = getopt.getopt(sys.argv[1:], "p:d:u:", ['period=', 'database=', 'url='])

        arg_period = Aggregator.REPORT_PERIOD
        arg_database_name = Aggregator.DATABASE
        arg_database_url = Aggregator.DATABASE_URL

        # Apply parameters if given
        for opt, arg in opts:
            if opt in ('-p', '--period'):
                arg_period = int(arg)
            elif opt in ('-d', '--database'):
                arg_database_name = arg
            elif opt in ('-u', '--url'):
                arg_database_url = arg

        Aggregator(database_name=arg_database_name, database_url=arg_database_url, report_period=arg_period, logger=log).run()

    # log.info the error messages in case of a parse error
    except getopt.GetoptError as err:
        log.info(err)
        log.info('Parse error; run the script using the following format: python aggregator.py -p <seconds> -d <database name> -u <database url>')
